@@ -28,7 +28,14 @@
 #define BACKEND_H
 
 #include <sys/queue.h>
-#include <pcre.h>
+
+#if !defined(HAVE_LIBPCRE2_8)
+#error "sniproxy requires libpcre2-8"
+#endif
+
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+
 #include "address.h"
 
 STAILQ_HEAD(Backend_head, Backend);
@@ -39,17 +46,25 @@ struct Backend {
     int use_proxy_header;
 
     /* Runtime fields */
-    pcre *pattern_re;
+    pcre2_code *pattern_re;
+    pcre2_match_data *pattern_match_data;
+    char *last_lookup_name;
+    size_t last_lookup_len;
+    size_t last_lookup_capacity;
+    int last_lookup_result;
+    int last_lookup_valid;
     STAILQ_ENTRY(Backend) entries;
 };
 
 void add_backend(struct Backend_head *, struct Backend *);
 int init_backend(struct Backend *);
+int valid_backend(const struct Backend *);
 struct Backend *lookup_backend(const struct Backend_head *, const char *, size_t);
 void print_backend_config(FILE *, const struct Backend *);
 void remove_backend(struct Backend_head *, struct Backend *);
-struct Backend *new_backend();
+struct Backend *new_backend(void);
 int accept_backend_arg(struct Backend *, const char *);
+void free_backend(struct Backend *);
 
 
 #endif
