@@ -26,12 +26,15 @@
 #ifndef CONNECTION_H
 #define CONNECTION_H
 
+#include <stddef.h>
 #include <sys/socket.h>
 #include <sys/queue.h>
 #include <ev.h>
 #include "listener.h"
 #include "buffer.h"
 
+struct DnsClientUsageEntry;
+
 struct Connection {
     enum State {
         NEW,            /* Before successful accept */
@@ -51,20 +54,40 @@ struct Connection {
         struct ev_io watcher;
         struct Buffer *buffer;
     } client, server;
+    struct ev_timer idle_timer;
+    struct ev_timer header_timer;
     struct Listener *listener;
     const char *hostname; /* Requested hostname */
     size_t hostname_len;
     size_t header_len;
     struct ResolvQuery *query_handle;
+    int dns_query_acquired;
+    struct DnsClientUsageEntry *dns_client_usage;
     ev_tstamp established_timestamp;
     int use_proxy_header;
 
     TAILQ_ENTRY(Connection) entries;
+    TAILQ_ENTRY(Connection) shrink_entries;
+    int shrink_candidate;
+    ev_tstamp shrink_deadline;
 };
 
-void init_connections();
+void init_connections(void);
 int accept_connection(struct Listener *, struct ev_loop *);
 void free_connections(struct ev_loop *);
-void print_connections();
+void print_connections(void);
+size_t connections_memory_usage_bytes(void);
+size_t connections_memory_peak_bytes(void);
+size_t connections_active_count(void);
+size_t connections_peak_count(void);
+
+#define DEFAULT_DNS_QUERY_CONCURRENCY 512
+#define DEFAULT_DNS_QUERIES_PER_CLIENT 16
+
+void connections_set_per_ip_connection_rate(double rate);
+void connections_set_dns_query_limit(size_t limit);
+void connections_set_dns_query_per_client_limit(size_t limit);
+void connections_set_buffer_limits(size_t client_limit, size_t server_limit);
+void connections_set_global_limit(size_t limit);
 
 #endif
