@@ -38,6 +38,7 @@
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <netinet/in.h>
+#include <stdint.h>
 #include <arpa/inet.h>
 #include <assert.h>
 #include "address.h"
@@ -47,6 +48,11 @@
 #include "protocol.h"
 #include "tls.h"
 #include "http.h"
+#include "fd_util.h"
+
+#define LISTENER_ACCEPT_MAX_BATCH 64
+/* Yield to the event loop after this many successful accepts to avoid starvation. */
+#define LISTENER_ACCEPT_YIELD_BATCH 16
 
 static void close_listener(struct ev_loop *, struct Listener *);
 static void accept_cb(struct ev_loop *, struct ev_io *, int);
@@ -57,6 +63,14 @@ static void free_listener(struct Listener *);
 static int parse_boolean(const char *);
 
 
+static void listener_acl_clear(struct Listener *);
+static void listener_acl_move(struct Listener *, struct Listener *);
+static int listener_acl_contains(const struct Listener *, const struct sockaddr_storage *);
+static int listener_acl_rule_match_v4(const struct ListenerACLRule *, const struct in_addr *);
+static int listener_acl_rule_match_v6(const struct ListenerACLRule *, const struct in6_addr *);
+static const char *listener_acl_rule_to_string(const struct ListenerACLRule *, char *, size_t);
+
+
 static int
 parse_boolean(const char *boolean) {
     const char *boolean_true[] = {
@@ -84,21 +98,163 @@ parse_boolean(const char *boolean) {
     return -1;
 }
 
+static void
+listener_acl_clear(struct Listener *listener) {
+    struct ListenerACLRule *rule;
+
+    if (listener == NULL)
+        return;
+
+    while ((rule = SLIST_FIRST(&listener->acl_rules)) != NULL) {
+        SLIST_REMOVE_HEAD(&listener->acl_rules, entries);
+        free(rule);
+    }
+
+    listener->acl_mode = LISTENER_ACL_MODE_DISABLED;
+}
+
+static void
+listener_acl_move(struct Listener *dst, struct Listener *src) {
+    if (dst == NULL || src == NULL)
+        return;
+
+    listener_acl_clear(dst);
+    dst->acl_mode = src->acl_mode;
+    dst->acl_rules = src->acl_rules;
+    SLIST_INIT(&src->acl_rules);
+    src->acl_mode = LISTENER_ACL_MODE_DISABLED;
+}
+
+static int
+listener_acl_rule_match_v4(const struct ListenerACLRule *rule, const struct in_addr *addr) {
+    if (rule == NULL || addr == NULL)
+        return 0;
+
+    if (rule->prefix_len == 0)
+        return 1;
+
+    uint32_t mask = rule->prefix_len == 32 ? UINT32_MAX : (~0u << (32 - rule->prefix_len));
+    uint32_t addr_val = ntohl(addr->s_addr);
+    uint32_t net_val = ntohl(rule->network.in.s_addr);
+
+    return (addr_val & mask) == (net_val & mask);
+}
+
+static int
+listener_acl_rule_match_v6(const struct ListenerACLRule *rule, const struct in6_addr *addr) {
+    if (rule == NULL || addr == NULL)
+        return 0;
+
+    if (rule->prefix_len == 0)
+        return 1;
+
+    unsigned full_bytes = rule->prefix_len / 8;
+    unsigned remaining_bits = rule->prefix_len % 8;
+
+    if (full_bytes > 0) {
+        if (memcmp(addr->s6_addr, rule->network.in6.s6_addr, full_bytes) != 0)
+            return 0;
+    }
+
+    if (remaining_bits > 0 && full_bytes < sizeof(rule->network.in6.s6_addr)) {
+        uint8_t mask = (uint8_t)(0xFF << (8 - remaining_bits));
+        if ((addr->s6_addr[full_bytes] & mask) !=
+                (rule->network.in6.s6_addr[full_bytes] & mask))
+            return 0;
+    }
+
+    return 1;
+}
+
+static int
+listener_acl_contains(const struct Listener *listener, const struct sockaddr_storage *addr) {
+    const struct ListenerACLRule *rule;
+
+    if (listener == NULL || addr == NULL)
+        return 0;
+
+    if (addr->ss_family == AF_INET) {
+        const struct sockaddr_in *sin = (const struct sockaddr_in *)addr;
+        SLIST_FOREACH(rule, &listener->acl_rules, entries) {
+            if (rule->family == AF_INET &&
+                    listener_acl_rule_match_v4(rule, &sin->sin_addr))
+                return 1;
+        }
+    } else if (addr->ss_family == AF_INET6) {
+        const struct sockaddr_in6 *sin6 = (const struct sockaddr_in6 *)addr;
+        if (IN6_IS_ADDR_V4MAPPED(&sin6->sin6_addr)) {
+            struct in_addr v4;
+            memcpy(&v4.s_addr, &sin6->sin6_addr.s6_addr[12], sizeof(v4.s_addr));
+            SLIST_FOREACH(rule, &listener->acl_rules, entries) {
+                if (rule->family == AF_INET &&
+                        listener_acl_rule_match_v4(rule, &v4))
+                    return 1;
+            }
+        } else {
+            SLIST_FOREACH(rule, &listener->acl_rules, entries) {
+                if (rule->family == AF_INET6 &&
+                        listener_acl_rule_match_v6(rule, &sin6->sin6_addr))
+                    return 1;
+            }
+        }
+    }
+
+    return 0;
+}
+
+static const char *
+listener_acl_rule_to_string(const struct ListenerACLRule *rule, char *buffer, size_t len) {
+    const void *addr = NULL;
+
+    if (buffer == NULL || len == 0)
+        return "(invalid)";
+
+    if (rule == NULL) {
+        snprintf(buffer, len, "(invalid)");
+        return buffer;
+    }
+
+    if (rule->family == AF_INET)
+        addr = &rule->network.in;
+    else if (rule->family == AF_INET6)
+        addr = &rule->network.in6;
+    else {
+        snprintf(buffer, len, "(invalid)");
+        return buffer;
+    }
+
+    if (inet_ntop(rule->family, addr, buffer, len) == NULL) {
+        snprintf(buffer, len, "(invalid)");
+        return buffer;
+    }
+
+    size_t used = strlen(buffer);
+    if (used + 5 >= len) {
+        snprintf(buffer, len, "(invalid)");
+        return buffer;
+    }
+
+    snprintf(buffer + used, len - used, "/%u", rule->prefix_len);
+
+    return buffer;
+}
+
 /*
  * Initialize each listener.
  */
 void
 init_listeners(struct Listener_head *listeners,
         const struct Table_head *tables, struct ev_loop *loop) {
-    struct Listener *iter;
+    struct Listener *iter = SLIST_FIRST(listeners);
     char address[ADDRESS_BUFFER_SIZE];
 
-    SLIST_FOREACH(iter, listeners, entries) {
+    while (iter != NULL) {
         if (init_listener(iter, tables, loop) < 0) {
             err("Failed to initialize listener %s",
                     display_address(iter->address, address, sizeof(address)));
             exit(1);
         }
+        iter = SLIST_NEXT(iter, entries);
     }
 }
 
@@ -168,24 +324,37 @@ listener_update(struct Listener *existing_listener, struct Listener *new_listene
     assert(new_listener != NULL);
     assert(address_compare(existing_listener->address, new_listener->address) == 0);
 
-    free(existing_listener->fallback_address);
-    existing_listener->fallback_address = new_listener->fallback_address;
-    new_listener->fallback_address = NULL;
+    if (existing_listener->fallback_address != new_listener->fallback_address) {
+        free(existing_listener->fallback_address);
+        existing_listener->fallback_address = new_listener->fallback_address;
+        new_listener->fallback_address = NULL;
+    }
+    existing_listener->fallback_use_proxy_header =
+            new_listener->fallback_use_proxy_header;
 
-    free(existing_listener->source_address);
-    existing_listener->source_address = new_listener->source_address;
-    new_listener->source_address = NULL;
+    if (existing_listener->source_address != new_listener->source_address) {
+        free(existing_listener->source_address);
+        existing_listener->source_address = new_listener->source_address;
+        new_listener->source_address = NULL;
+    }
+    existing_listener->transparent_proxy = new_listener->transparent_proxy;
 
     existing_listener->protocol = new_listener->protocol;
 
-    free(existing_listener->table_name);
-    existing_listener->table_name = new_listener->table_name;
-    new_listener->table_name = NULL;
+    if (existing_listener->table_name != new_listener->table_name) {
+        free(existing_listener->table_name);
+        existing_listener->table_name = new_listener->table_name;
+        new_listener->table_name = NULL;
+    }
 
     logger_ref_put(existing_listener->access_log);
     existing_listener->access_log = logger_ref_get(new_listener->access_log);
 
     existing_listener->log_bad_requests = new_listener->log_bad_requests;
+    existing_listener->reuseport = new_listener->reuseport;
+    existing_listener->ipv6_v6only = new_listener->ipv6_v6only;
+
+    listener_acl_move(existing_listener, new_listener);
 
     struct Table *new_table =
             table_lookup(tables, existing_listener->table_name);
@@ -202,7 +371,7 @@ listener_update(struct Listener *existing_listener, struct Listener *new_listene
 }
 
 struct Listener *
-new_listener() {
+new_listener(void) {
     struct Listener *listener = calloc(1, sizeof(struct Listener));
     if (listener == NULL) {
         err("calloc");
@@ -220,6 +389,8 @@ new_listener() {
     listener->ipv6_v6only = 0;
     listener->transparent_proxy = 0;
     listener->fallback_use_proxy_header = 0;
+    listener->acl_mode = LISTENER_ACL_MODE_DISABLED;
+    SLIST_INIT(&listener->acl_rules);
     listener->reference_count = 0;
     /* Initializes sock fd to negative sentinel value to indicate watchers
      * are not active */
@@ -270,6 +441,7 @@ accept_listener_table_name(struct Listener *listener, const char *table_name) {
         err("Duplicate table: %s", table_name);
         return 0;
     }
+    free(listener->table_name);
     listener->table_name = strdup(table_name);
     if (listener->table_name == NULL) {
         err("%s: strdup", __func__);
@@ -337,15 +509,9 @@ accept_listener_fallback_address(struct Listener *listener, const char *fallback
             listener->fallback_address = fallback_address;
             return 1;
         } else if (address_is_hostname(fallback_address)) {
-#ifndef HAVE_LIBUDNS
-            err("Only fallback socket addresses permitted when compiled without libudns");
-            free(fallback_address);
-            return 0;
-#else
             warn("Using hostname as fallback address is strongly discouraged");
             listener->fallback_address = fallback_address;
             return 1;
-#endif
         } else if (address_is_wildcard(fallback_address)) {
             /* The wildcard functionality requires successfully parsing the
              * hostname from the client's request, if we couldn't find the
@@ -492,6 +658,8 @@ valid_listener(const struct Listener *listener) {
 static int
 init_listener(struct Listener *listener, const struct Table_head *tables,
         struct ev_loop *loop) {
+    int sockfd = -1;
+    int rc = -1;
     char address[ADDRESS_BUFFER_SIZE];
     struct Table *table = table_lookup(tables, listener->table_name);
     if (table == NULL) {
@@ -508,14 +676,24 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
         address_set_port(listener->fallback_address,
                 address_port(listener->address));
 
+    int socket_type = SOCK_STREAM;
 #ifdef HAVE_ACCEPT4
-    int sockfd = socket(address_sa(listener->address)->sa_family, SOCK_STREAM | SOCK_NONBLOCK, 0);
-#else
-    int sockfd = socket(address_sa(listener->address)->sa_family, SOCK_STREAM, 0);
+    socket_type |= SOCK_NONBLOCK;
 #endif
+#ifdef SOCK_CLOEXEC
+    socket_type |= SOCK_CLOEXEC;
+#endif
+    sockfd = socket(address_sa(listener->address)->sa_family, socket_type, 0);
     if (sockfd < 0) {
         err("socket failed: %s", strerror(errno));
-        return sockfd;
+        rc = sockfd;
+        goto error;
+    }
+
+    if (set_cloexec(sockfd) < 0) {
+        err("failed to set close-on-exec on listener socket: %s", strerror(errno));
+        rc = -1;
+        goto error;
     }
 
     /* set SO_REUSEADDR on server socket to facilitate restart */
@@ -523,8 +701,8 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
     int result = setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
     if (result < 0) {
         err("setsockopt SO_REUSEADDR failed: %s", strerror(errno));
-        close(sockfd);
-        return result;
+        rc = result;
+        goto error;
     }
 
     /* set SO_KEEPALIVE on the server socket so that abandoned client connections
@@ -532,8 +710,8 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
     result = setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on));
     if (result < 0) {
         err("setsockopt SO_KEEPALIVE failed: %s", strerror(errno));
-        close(sockfd);
-        return result;
+        rc = result;
+        goto error;
     }
 
     if (listener->reuseport == 1) {
@@ -546,8 +724,8 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
 #endif
         if (result < 0) {
             err("setsockopt SO_REUSEPORT failed: %s", strerror(errno));
-            close(sockfd);
-            return result;
+            rc = result;
+            goto error;
         }
     }
 
@@ -562,8 +740,8 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
 #endif
         if (result < 0) {
             err("setsockopt IPV6_V6ONLY failed: %s", strerror(errno));
-            close(sockfd);
-            return result;
+            rc = result;
+            goto error;
         }
     }
 
@@ -577,21 +755,28 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
         if (sockfd < 0) {
             err("binder failed to bind to %s",
                 display_address(listener->address, address, sizeof(address)));
-            return sockfd;
+            rc = sockfd;
+            goto error;
+        }
+
+        if (set_cloexec(sockfd) < 0) {
+            err("failed to set close-on-exec on bound listener socket: %s", strerror(errno));
+            rc = -1;
+            goto error;
         }
     } else if (result < 0) {
         err("bind %s failed: %s",
             display_address(listener->address, address, sizeof(address)),
             strerror(errno));
-        close(sockfd);
-        return result;
+        rc = result;
+        goto error;
     }
 
     result = listen(sockfd, SOMAXCONN);
     if (result < 0) {
         err("listen failed: %s", strerror(errno));
-        close(sockfd);
-        return result;
+        rc = result;
+        goto error;
     }
 
 #ifndef HAVE_ACCEPT4
@@ -606,6 +791,15 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
     ev_io_start(loop, &listener->watcher);
 
     return sockfd;
+
+error:
+    if (sockfd >= 0)
+        close(sockfd);
+    if (listener->table != NULL) {
+        table_ref_put(listener->table);
+        listener->table = NULL;
+    }
+    return rc;
 }
 
 /*
@@ -618,6 +812,23 @@ init_listener(struct Listener *listener, const struct Table_head *tables,
 struct LookupResult
 listener_lookup_server_address(const struct Listener *listener,
         const char *name, size_t name_len) {
+    if (listener == NULL)
+        return (struct LookupResult){ .address = NULL };
+
+    if (name == NULL || name_len == 0) {
+        return (struct LookupResult){
+            .address = listener->fallback_address,
+            .use_proxy_header = listener->fallback_use_proxy_header,
+        };
+    }
+
+    if (listener->table == NULL) {
+        return (struct LookupResult){
+            .address = listener->fallback_address,
+            .use_proxy_header = listener->fallback_use_proxy_header,
+        };
+    }
+
     struct LookupResult table_result =
         table_lookup_server_address(listener->table, name, name_len);
 
@@ -666,6 +877,18 @@ listener_lookup_server_address(const struct Listener *listener,
          * port from the listen, this allows sharing table across listeners */
         struct Address *new_addr = copy_address(table_result.address);
 
+        if (new_addr == NULL) {
+            err("Failed to copy backend address for %.*s: %s",
+                    (int)name_len,
+                    name,
+                    strerror(errno));
+
+            return (struct LookupResult){
+                .address = listener->fallback_address,
+                .use_proxy_header = listener->fallback_use_proxy_header
+            };
+        }
+
         address_set_port(new_addr, address_port(listener->address));
 
         return (struct LookupResult){
@@ -678,9 +901,32 @@ listener_lookup_server_address(const struct Listener *listener,
     }
 }
 
+int
+listener_acl_allows(const struct Listener *listener,
+        const struct sockaddr_storage *addr) {
+    if (listener == NULL)
+        return 0;
+
+    if (listener->acl_mode == LISTENER_ACL_MODE_DISABLED)
+        return 1;
+
+    int matched = listener_acl_contains(listener, addr);
+
+    if (listener->acl_mode == LISTENER_ACL_MODE_ALLOW_EXCEPT)
+        return matched ? 0 : 1;
+
+    if (listener->acl_mode == LISTENER_ACL_MODE_DENY_EXCEPT)
+        return matched ? 1 : 0;
+
+    return 1;
+}
+
 void
 print_listener_config(FILE *file, const struct Listener *listener) {
+    assert(listener != NULL);
+
     char address[ADDRESS_BUFFER_SIZE];
+    const struct ListenerACLRule *rule;
 
     fprintf(file, "listener %s {\n",
             display_address(listener->address, address, sizeof(address)));
@@ -710,6 +956,19 @@ print_listener_config(FILE *file, const struct Listener *listener) {
     if (listener->reuseport)
         fprintf(file, "\treuseport on\n");
 
+    if (listener->acl_mode != LISTENER_ACL_MODE_DISABLED) {
+        const char *policy = listener->acl_mode == LISTENER_ACL_MODE_ALLOW_EXCEPT ?
+                "allow_except" : "deny_except";
+        fprintf(file, "\tacl %s {\n", policy);
+        SLIST_FOREACH(rule, &listener->acl_rules, entries) {
+            char cidr_buf[INET6_ADDRSTRLEN + 8];
+            const char *cidr = listener_acl_rule_to_string(rule, cidr_buf, sizeof(cidr_buf));
+            if (cidr != NULL)
+                fprintf(file, "\t\t%s\n", cidr);
+        }
+        fprintf(file, "\t}\n");
+    }
+
     fprintf(file, "}\n\n");
 }
 
@@ -729,6 +988,8 @@ free_listener(struct Listener *listener) {
     if (listener == NULL)
         return;
 
+    listener_acl_clear(listener);
+
     free(listener->address);
     free(listener->fallback_address);
     free(listener->source_address);
@@ -768,7 +1029,10 @@ listener_ref_put(struct Listener *listener) {
     if (listener == NULL)
         return;
 
-    assert(listener->reference_count > 0);
+    if (listener->reference_count == 0) {
+        err("%s: reference_count underflow", __func__);
+        return;
+    }
     listener->reference_count--;
     if (listener->reference_count == 0)
         free_listener(listener);
@@ -776,29 +1040,67 @@ listener_ref_put(struct Listener *listener) {
 
 struct Listener *
 listener_ref_get(struct Listener *listener) {
+    assert(listener != NULL);
+
     listener->reference_count++;
     return listener;
 }
 
+void
+cleanup_listener(void *listener_ptr) {
+    struct Listener *listener = (struct Listener *)listener_ptr;
+    if (listener == NULL)
+        return;
+
+    /* Use reference counting to properly free listener and its resources */
+    listener_ref_get(listener);
+    listener_ref_put(listener);
+}
+
 static void
 accept_cb(struct ev_loop *loop, struct ev_io *w, int revents) {
     struct Listener *listener = (struct Listener *)w->data;
 
-    if (revents & EV_READ) {
+    if (!(revents & EV_READ))
+        return;
+
+    int last_errno = 0;
+    int accepted = 0;
+
+    for (int i = 0; i < LISTENER_ACCEPT_MAX_BATCH; i++) {
         int result = listener->accept_cb(listener, loop);
-        if (result == 0 && (errno == EMFILE || errno == ENFILE)) {
-            char address_buf[ADDRESS_BUFFER_SIZE];
-            int backoff_time = 2;
-
-            err("File descriptor limit reached! "
-                "Suspending accepting new connections on %s for %d seconds",
-                display_address(listener->address, address_buf, sizeof(address_buf)),
-                backoff_time);
-            ev_io_stop(loop, w);
-
-            ev_timer_set(&listener->backoff_timer, backoff_time, 0.0);
-            ev_timer_start(loop, &listener->backoff_timer);
+        if (result > 0) {
+            last_errno = 0;
+            accepted++;
+            if (accepted >= LISTENER_ACCEPT_YIELD_BATCH)
+                break;
+            continue;
         }
+
+        last_errno = errno;
+
+        if (last_errno == EINTR)
+            continue;
+
+        if (last_errno == EAGAIN || last_errno == EWOULDBLOCK ||
+                last_errno == 0)
+            break;
+
+        break;
+    }
+
+    if (last_errno == EMFILE || last_errno == ENFILE) {
+        char address_buf[ADDRESS_BUFFER_SIZE];
+        int backoff_time = 2;
+
+        err("File descriptor limit reached! "
+            "Suspending accepting new connections on %s for %d seconds",
+            display_address(listener->address, address_buf, sizeof(address_buf)),
+            backoff_time);
+        ev_io_stop(loop, w);
+
+        ev_timer_set(&listener->backoff_timer, backoff_time, 0.0);
+        ev_timer_start(loop, &listener->backoff_timer);
     }
 }
 
