@@ -37,13 +37,20 @@ struct Logger;
 #define LOG_INFO    6
 #define LOG_DEBUG   7
 
+#include <sys/types.h>
+
 struct Logger *new_syslog_logger(const char *facility);
 struct Logger *new_file_logger(const char *filepath);
 void set_default_logger(struct Logger *);
 void set_logger_priority(struct Logger *, int);
 struct Logger *logger_ref_get(struct Logger *);
 void logger_ref_put(struct Logger *);
-void reopen_loggers();
+void reopen_loggers(void);
+void logger_for_each_file_sink(void (*callback)(const char *, void *), void *userdata);
+void logger_prepare_process_title(int argc, char **argv);
+int logger_process_is_active(void);
+void logger_parent_notify_fs_locked(void);
+int logger_drop_privileges(uid_t uid, gid_t gid);
 
 /* Shorthand to log to global error log */
 void fatal(const char *, ...)
@@ -63,4 +70,8 @@ void debug(const char *, ...)
 void log_msg(struct Logger *, int, const char *, ...)
     __attribute__ ((format (printf, 3, 4)));
 
+/* Global resolver debug flag */
+void set_resolver_debug(int enabled);
+int get_resolver_debug(void);
+
 #endif
