@@ -27,14 +27,26 @@
 #define TABLE_H
 
 #include <stdio.h>
+#include <stdint.h>
 #include <sys/queue.h>
 #include "backend.h"
 #include "address.h"
 
 #define TABLE_NAME_LEN 20
+#define TABLE_CACHE_MAX_NAME_LEN 255
+#define TABLE_CACHE_BITS 8
+#define TABLE_CACHE_SIZE (1u << TABLE_CACHE_BITS)
 
 SLIST_HEAD(Table_head, Table);
 
+struct TableCacheEntry {
+    uint32_t hash;
+    uint16_t name_len;
+    uint16_t generation;
+    char name[TABLE_CACHE_MAX_NAME_LEN + 1];
+    struct Backend *backend;
+};
+
 struct Table {
     char *name;
     int use_proxy_header;
@@ -42,6 +54,8 @@ struct Table {
     /* Runtime fields */
     int reference_count;
     struct Backend_head backends;
+    uint16_t cache_generation;
+    struct TableCacheEntry cache[TABLE_CACHE_SIZE];
     SLIST_ENTRY(Table) entries;
 };
 
@@ -51,7 +65,7 @@ struct LookupResult {
     int use_proxy_header;
 };
 
-struct Table *new_table();
+struct Table *new_table(void);
 int accept_table_arg(struct Table *, const char *);
 void add_table(struct Table_head *, struct Table *);
 struct Table *table_lookup(const struct Table_head *, const char *);
@@ -63,6 +77,7 @@ int valid_table(struct Table *);
 void init_table(struct Table *);
 void table_ref_put(struct Table *);
 struct Table *table_ref_get(struct Table *);
+void cleanup_table(void *);
 void tables_reload(struct Table_head *, struct Table_head *);
 
 void free_tables(struct Table_head *);
