@@ -1,3 +1,29 @@
+/*
+ * Copyright (c) 2025, Renaud Allard <renaud@allard.it>
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
+ */
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -6,8 +32,8 @@
 #include "address.h"
 
 struct Test {
-    char *input;
-    char *output;
+    const char *input;
+    const char *output;
     int expected_type;
     int port;
 };
@@ -47,7 +73,12 @@ static const char *bad[] = {
     NULL,
     "www..example.com",
     "-www.example.com",
-    "1n\\/l1|>|-|0$T|\\|4M"
+    "1n\\/l1|>|-|0$T|\\|4M",
+    "example.com:1e3",
+    "example.com:999999999999999999999999",
+    "example.com:+443",
+    "example.com:-1",
+    "example.com: 80"
 };
 
 int compare_address_strings(const char *a, const char *b) {
@@ -62,7 +93,7 @@ int compare_address_strings(const char *a, const char *b) {
     return result;
 }
 
-int main() {
+int main(void) {
     /* using volatile variables so we can examine core dumps */
     for (volatile unsigned int i = 0; i < sizeof(good) / sizeof(struct Test); i++) {
         int port;
@@ -105,7 +136,9 @@ int main() {
             return 1;
         }
 
-        assert(display_address(addr, NULL, 0) == NULL);
+        const char *invalid_display = display_address(addr, NULL, 0);
+        assert(invalid_display != NULL);
+        assert(strcmp(invalid_display, "(invalid)") == 0);
 
         port = address_port(addr);
 
@@ -166,4 +199,3 @@ int main() {
 
     return 0;
 }
-
